#include <stdlib.h>

#include "doomdef.h"

#include "i_system.h"
#include "z_zone.h"
#include "w_wad.h"

#include "doomdef.h"
#include "doomstat.h"

#include "r_local.h"
#include "r_sky.h"

#include "p_local.h"

#include "kg_3dfloor.h"

planefunction_t		floorfunc;
planefunction_t		ceilingfunc;

//
// opening
//

// Here comes the obnoxious "visplane".
#define MAXVISPLANES	1024
visplane_t		visplanes[MAXVISPLANES];
visplane_t*		lastvisplane;
visplane_t*		floorplane;
visplane_t*		ceilingplane;

// ?
#define MAXOPENINGS	SCREENWIDTH*64
short			openings[MAXOPENINGS];
short*			lastopening;


//
// Clip values are the solid pixel bounding the range.
//  floorclip starts out SCREENHEIGHT
//  ceilingclip starts out -1
//
short			floorclip[SCREENWIDTH];
short			ceilingclip[SCREENWIDTH];

//
// spanstart holds the start of a plane span
// initialized to 0 at start
//
int			spanstart[SCREENHEIGHT];
int			spanstop[SCREENHEIGHT];

//
// texture mapping
//
uint8_t*		planezlight;
fixed_t			planeheight;
int			planexlight;

fixed_t			yslope[SCREENHEIGHT];
fixed_t			distscale[SCREENWIDTH];
fixed_t			basexscale;
fixed_t			baseyscale;

fixed_t			cachedheight[SCREENHEIGHT];
fixed_t			cacheddistance[SCREENHEIGHT];
fixed_t			cachedxstep[SCREENHEIGHT];
fixed_t			cachedystep[SCREENHEIGHT];

void *dc_fogmap;

boolean draw_fog_first;

void R_ClearPlane(visplane_t *pl)
{
	int i;

	for(i = 0; i < SCREENWIDTH; i++)
		pl->top[i] = 0x7FFF;
}

//
// R_InitPlanes
// Only at game startup.
//
void R_InitPlanes (void)
{
  // Doh!
}


//
// R_MapPlane
//
// Uses global vars:
//  planeheight
//  ds_source
//  basexscale
//  baseyscale
//  viewx
//  viewy
//
// BASIC PRIMITIVE
//
void
R_MapPlane
( int		y,
  int		x1,
  int		x2 )
{
    angle_t	angle;
    fixed_t	distance;
    fixed_t	length;
    unsigned	index;
	
#ifdef RANGECHECK
    if (x2 < x1
	|| x1<0
	|| x2>=viewwidth
	|| (unsigned)y>viewheight)
    {
	I_Error ("R_MapPlane: %i, %i at %i",x1,x2,y);
    }
#endif

    if (planeheight != cachedheight[y])
    {
	cachedheight[y] = planeheight;
	distance = cacheddistance[y] = FixedMul (planeheight, yslope[y]);
	ds_xstep = cachedxstep[y] = FixedMul (distance,basexscale);
	ds_ystep = cachedystep[y] = FixedMul (distance,baseyscale);
    }
    else
    {
	distance = cacheddistance[y];
	ds_xstep = cachedxstep[y];
	ds_ystep = cachedystep[y];
    }
	
    length = FixedMul (distance,distscale[x1]);
    angle = (viewangle + xtoviewangle[x1])>>ANGLETOFINESHIFT;
    ds_xfrac = viewx + FixedMul(finecosine[angle], length);
    ds_yfrac = -viewy - FixedMul(finesine[angle], length);

    if (fixedcolormap)
	dc_colormap = fixedcolormap;
    else
    {
	int shadelevel;

	index = distance >> LIGHTZSHIFT;

	if (index >= MAXLIGHTZ )
	    index = MAXLIGHTZ-1;

	shadelevel = planezlight[index] + planexlight;
	if(shadelevel < 0)
		shadelevel = 0;
	if(shadelevel >= LIGHTLEVELS)
		shadelevel = LIGHTLEVELS-1;
	dc_colormap = dc_fogmap + shadelevel * 256;
    }

    ds_y = y;
    ds_x1 = x1;
    ds_x2 = x2;

    // high or low detail
    if(draw_fog_first)
	R_DrawSpanFog();
    spanfunc ();	
}


//
// R_ClearPlanes
// At begining of frame.
//
void R_ClearPlanes (void)
{
    int		i;
    angle_t	angle;
    
    // opening / clipping determination
    for (i=0 ; i<viewwidth ; i++)
    {
	floorclip[i] = viewheight;
	ceilingclip[i] = -1;
    }

    lastvisplane = visplanes;
    lastopening = openings;
    
    // texture calculation
    memset (cachedheight, 0, sizeof(cachedheight));

    // left to right mapping
    angle = (viewangle-ANG90)>>ANGLETOFINESHIFT;
	
    // scale will be unit scale at SCREENWIDTH/2 distance
    basexscale = FixedDiv (finecosine[angle],centerxfrac);
    baseyscale = -FixedDiv (finesine[angle],centerxfrac);
}

extern render_t render_normal;

//
// R_FindPlane
//
visplane_t*
R_FindPlane
( fixed_t	height,
  int		picnum,
  int		lightlevel,
  void		*colormap,
  void		*fogmap,
  render_t	*render,
  boolean	has_fog)
{
    visplane_t*	check;
    int is3d = !!fakeplane;

    if (picnum == skyflatnum)
    {
	height = 0;			// all skys map together
	lightlevel = 0;
	colormap = colormaps;
	fogmap = NULL;
	render = &render_normal;
	has_fog = false;
    }

    for (check=visplanes; check<lastvisplane; check++)
    {
	if (height == check->height
	    && picnum == check->picnum
	    && lightlevel == check->lightlevel
	    // [kg] colormap
	    && colormap == check->colormap
	    // [kg] fogmap
	    && fogmap == check->fogmap
	    // [kg] 3D planes
	    && is3d == check->is3d
	    // [kg] render style
	    && render->renderstyle == check->render.renderstyle
	    && render->rendertable == check->render.rendertable
	    // [kg] fog boundary
	    && has_fog == check->hasFog
	)
	{
	    break;
	}
    }

    if (check < lastvisplane)
	return check;

    if (lastvisplane - visplanes == MAXVISPLANES)
	I_Error ("R_FindPlane: no more visplanes");

    lastvisplane++;

    check->height = height;
    check->picnum = picnum;
    check->lightlevel = lightlevel;
    check->colormap = colormap;
    check->fogmap = fogmap;
    check->render.renderstyle = render->renderstyle;
    check->render.rendertable = render->rendertable;
    check->minx = SCREENWIDTH;
    check->maxx = -1;
    check->is3d = is3d;
    check->hasFog = has_fog;

//    memset (check->top,0xff,sizeof(check->top));
    R_ClearPlane(check);

    return check;
}


//
// R_CheckPlane
//
visplane_t*
R_CheckPlane
( visplane_t*	pl,
  int		start,
  int		stop )
{
    int		intrl;
    int		intrh;
    int		unionl;
    int		unionh;
    int		x;
	
    if (start < pl->minx)
    {
	intrl = pl->minx;
	unionl = start;
    }
    else
    {
	unionl = pl->minx;
	intrl = start;
    }
	
    if (stop > pl->maxx)
    {
	intrh = pl->maxx;
	unionh = stop;
    }
    else
    {
	unionh = pl->maxx;
	intrh = stop;
    }

    for (x=intrl ; x<= intrh ; x++)
	if (pl->top[x] != 0x7fff)
	    break;

    if (x > intrh)
    {
	pl->minx = unionl;
	pl->maxx = unionh;

	// use the same one
	return pl;		
    }

    // make a new visplane
    lastvisplane->height = pl->height;
    lastvisplane->picnum = pl->picnum;
    lastvisplane->lightlevel = pl->lightlevel;
    lastvisplane->colormap = pl->colormap;
    lastvisplane->fogmap = pl->fogmap;
    lastvisplane->is3d = pl->is3d;
    lastvisplane->hasFog = pl->hasFog;
    lastvisplane->render.renderstyle = pl->render.renderstyle;
    lastvisplane->render.rendertable = pl->render.rendertable;

    if (lastvisplane - visplanes == MAXVISPLANES)
	I_Error ("R_FindPlane: no more visplanes");

    pl = lastvisplane++;
    pl->minx = start;
    pl->maxx = stop;

//    memset (pl->top,0xff,sizeof(pl->top));
    R_ClearPlane(pl);

    return pl;
}


//
// R_MakeSpans
//
void
R_MakeSpans
( int		x,
  int		t1,
  int		b1,
  int		t2,
  int		b2 )
{
    while (t1 < t2 && t1<=b1)
    {
	R_MapPlane (t1,spanstart[t1],x-1);
	t1++;
    }
    while (b1 > b2 && b1>=t1)
    {
	R_MapPlane (b1,spanstart[b1],x-1);
	b1--;
    }
	
    while (t2 < t1 && t2<=b2)
    {
	spanstart[t2] = x;
	t2++;
    }
    while (b2 > b1 && b2>=t2)
    {
	spanstart[b2] = x;
	b2--;
    }
}



//
// R_DrawPlanes
// At the end of each frame.
// [kg] draw solid and 3D separately
//
void R_DrawPlanes(fixed_t height)
{
    visplane_t	*pl;
    int		light;
    int		x;
    int		stop;
    int		angle;
    int		texture;
				
#ifdef RANGECHECK
    if (ds_p - drawsegs > MAXDRAWSEGS)
	I_Error ("R_DrawPlanes: drawsegs overflow (%i)",
		 ds_p - drawsegs);
    
    if (lastvisplane - visplanes > MAXVISPLANES)
	I_Error ("R_DrawPlanes: visplane overflow (%i)",
		 lastvisplane - visplanes);
    
    if (lastopening - openings > MAXOPENINGS)
	I_Error ("R_DrawPlanes: opening overflow (%i)",
		 lastopening - openings);
#endif

    for (pl = visplanes ; pl < lastvisplane ; pl++)
    {
	if (pl->minx > pl->maxx)
	    continue;

	if(!pl->picnum && !pl->hasFog)
	    continue;

	if(pl->is3d && pl->height != height)
	    continue;

	if(!pl->is3d && height != ONFLOORZ)
	    continue;

	// sky flat
	if (pl->picnum == skyflatnum)
	{
	    if(!skytexture)
		continue;

	    colfunc = R_DrawColumn;

	    dc_iscale = pspriteiscale;

	    dc_lightcolor = colormaps;
	    if(!fixedcolormap)
		dc_colormap = colormaps;
	    dc_texturemid = skytexturemid;
	    for (x=pl->minx ; x <= pl->maxx ; x++)
	    {
		dc_yl = pl->top[x];
		dc_yh = pl->bottom[x];

		if (dc_yl <= dc_yh)
		{
		    angle = (viewangle + xtoviewangle[x])>>ANGLETOSKYSHIFT;
		    dc_x = x;
		    dc_source = R_GetColumn(skytexture, angle);
		    dc_src_height = textures[skytexture].height;
		    colfunc ();
		}
	    }
	    continue;
	}

	// wall texture
	texture = texturetranslation[pl->picnum];
	dc_source = textures[texture].data;
	dc_src_height = textures[texture].width;
	dc_src_width = textures[texture].height;

	planeheight = abs(pl->height-viewz);
	light = ((pl->lightlevel & 0xFF) >> LIGHTSEGSHIFT) + extralight;

	if (light >= LIGHTLEVELS)
	    light = LIGHTLEVELS-1;

	if (light < 0)
	    light = 0;

	if(pl->lightlevel & 0x8000)
	    planezlight = zlight_alt[light];
	else
	    planezlight = zlight[light];
	planexlight = (pl->lightlevel >> 8) & 31;

	dc_lightcolor = pl->colormap;
	if(pl->fogmap)
		dc_fogmap = pl->fogmap;
	else
		dc_fogmap = colormaps;

	draw_fog_first = pl->hasFog;

	if(!pl->picnum)
	{
		if(!draw_fog_first)
			return;
		draw_fog_first = false;
		R_SetupRenderFuncSpan(RENDER_FOG_ONLY, NULL, NULL, pl->is3d);
	} else
		R_SetupRenderFuncSpan(pl->render.renderstyle, pl->render.rendertable, NULL, pl->is3d);

	pl->top[pl->maxx+1] = 0x7fff;
	pl->top[pl->minx-1] = 0x7fff;
		
	stop = pl->maxx + 1;

	for (x=pl->minx ; x<= stop ; x++)
	{
	    R_MakeSpans(x,pl->top[x-1],
			pl->bottom[x-1],
			pl->top[x],
			pl->bottom[x]);
	}
    }
}

